/*
 * Test dataspace operations.
 *
 * Copyright (C) 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <ipc/mem_ipc.h>

#include "dataspace_client.h"
#include "opener_client.h"
#include "opener_context_client.h"
#include "mapped_file_client.h"
#include "memory_utils.h"



int main(void)
{
  l4_cap_idx_t server = l4re_env_get_cap("server");
  l4_cap_idx_t context_ref;
  client_Opener opener(server);

  long err = opener.context(&context_ref);

  if (err)
  {
    printf("Could not obtain context: %s\n", l4sys_errtostr(err));
    return 1;
  }

  client_OpenerContext context(context_ref);
  client_Dataspace context_ds(context_ref);
  unsigned long size, flags;

  err = context_ds.info(&size, &flags);

  if (err)
  {
    printf("Could not obtain context info: %s\n", l4sys_errtostr(err));
    return 1;
  }

  /* Map context memory to write the filename. */

  char *memory;

  printf("Attach region of size %ld...\n", size);

  err = ipc_attach_dataspace(context_ref, size, (void **) &memory);

  if (err)
  {
    printf("Could not map memory: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Mapped memory at %p\n", memory);

  /* Write the filename. */

  strcpy(memory, "123");

  /* Invoke the open method to receive the file reference. */

  l4_cap_idx_t file;

  err = context.open(L4_FPAGE_RW, &size, &file);

  if (err)
  {
    printf("Could not obtain file: %s\n", l4sys_errtostr(err));
    return 1;
  }

  client_MappedFile obj(file);

  /* Some memory to be mapped. */

  size_t start_pos, end_pos, data_end;

  printf("Map region from %ld to %ld...\n", 0L, page(10));

  err = obj.mmap(0, page(10), &start_pos, &end_pos, &data_end);

  if (err)
  {
    printf("Could not map file region: %s\n", l4sys_errtostr(err));
    return 1;
  }

  size = end_pos - start_pos;

  printf("Attach region of size %ld...\n", size);

  err = ipc_attach_dataspace(file, size, (void **) &memory);

  if (err)
  {
    printf("Could not map memory: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Mapped memory at %p\n", memory);

  for (unsigned long offset = 0; offset < size; offset += 1024)
  {
    char buf[11];
    printf("10 bytes from %p...\n", (memory + offset));
    strncpy(buf, (memory + offset), 10);
    buf[10] = '\0';
    printf("%s\n", buf);
  }

  return 0;
}
